//
// This is a self-contained source code of a shared library that provides numerical solutions 
// of the diffusion-degradation equation in 3D. The library functions can be invoked by Comsol 
// Multiphysics as external functions.
//

#include <cmath>
#include <random>
#include <algorithm>
#include <cstdlib>
#include <cstring> // strcmp
#include <limits>
#ifdef LOGGING_ENABLED
#  include <fstream>
#  include <iostream>
#endif

using std::sqrt;


// == Global variables ============================================================================

// -- API data compliance -----------------------------------------------------
//
int const COMSOL_EXIT_SUCCESS = 1;
int const COMSOL_EXIT_FAILURE = 0;
//
static const char* error = NULL;

// -- precomputed values ------------------------------------------------------
//
double constexpr pi = 4.*std::atan(1);
double constexpr sqrt_pi = sqrt(pi);
double constexpr pi_pow_3over2 = pi*sqrt_pi;
double constexpr nearly_zero = std::numeric_limits<double>::min();

// -- settings and data -------------------------------------------------------
//
#define N_SOURCES (4*24)   // per whole simulation
//
unsigned int const RANDOM_SEED = 123456u;
double const CELL_RADIUS = 1.0/M_PI;
double const MAX_SOURCE_DIST_FROM_ORIGIN = 10.0*CELL_RADIUS;
//
double const SOURCES_ACTIVITY_DURATION   = 24*60*60;  // 24 h
double const SINGLE_SOURCE_MEAN_DURATION =    20*60.; // 20 min
//
double const SOURCES_ALL_DIFFUSIVITY = 1e-3;        // 1e-3 == 10*D_cyt == 100*D_mem
double const SOURCES_ALL_DEGRADATION = 1./(20*60.); // 1/(20 min)
//
static double sources_positions_ [N_SOURCES][3]; // x, y, z
static double sources_activities_[N_SOURCES][3]; // time on, time off, amplitude

double const SOURCES_AMPLITUDE_LOGNORM_MEAN   = -1.25;
double const SOURCES_AMPLITUDE_LOGNORM_STDDEV =  0.5;



// == Functions' declarations =====================================================================

// -- API-compliant functions
//
extern "C" int init(const char*) ;
extern "C" const char* getLastError() ;
extern "C" int eval(const char* func, int nArgs, const double** inReal, const double** inImag,
                    int blockSize, double* outReal, double* outImag) ;

// -- auxiliary functions 
//
void randomize_sources_positions_(void);
void randomize_sources_activities_(void);
double kernel_   (double x, double y, double z, double t, double S, double D, double G) ;
double kernel_dx_(double x, double y, double z, double t, double S, double D, double G) ;
double kernel_dy_(double x, double y, double z, double t, double S, double D, double G) ;
double kernel_dz_(double x, double y, double z, double t, double S, double D, double G) ;
double kernel_dt_(double x, double y, double z, double t, double S, double D, double G) ;
double kernel_dcoord__(double dc,double x,double y,double z,double t,double S,double D,double G);
template<class TKernelF>double sum_kernels_(TKernelF k, double x, double y, double z, double t) ;



// == Functions' definitions ======================================================================

// -- API-compliant functions -------------------------------------------------


extern "C" int init(const char *str) 
{
  randomize_sources_positions_();
  randomize_sources_activities_();
  return COMSOL_EXIT_SUCCESS;
}


extern "C" const char* getLastError() 
{
  return error;
}


extern "C" int eval(const char* func, int nArgs, const double** inReal, const double** inImag, 
                    int blockSize, double* outReal, double* outImag) 
{
  if (nArgs != 4) {
    error = "LIBEGF: eval(): Four arguments expected.";
    return COMSOL_EXIT_FAILURE;
  }

  auto const kernel = (!strcmp("egf",   func)) ? kernel_ 
                    : (!strcmp("egfdx", func)) ? kernel_dx_
                    : (!strcmp("egfdy", func)) ? kernel_dy_
                    : (!strcmp("egfdz", func)) ? kernel_dz_
                    : (!strcmp("egfdt", func)) ? kernel_dt_ 
                    : [](double,double,double,double,double,double,double) -> double
                      {  error = "LIBEGF: Unknown function.";   return -1.0;  } ;

  for (int i = 0; i < blockSize; i++) {
    double x = inReal[0][i];
    double y = inReal[1][i];
    double z = inReal[2][i];
    double t = inReal[3][i];
    outReal[i] = sum_kernels_(kernel, x,y,z, t);
  }

  return COMSOL_EXIT_SUCCESS;
}



// -- auxiliary functions -----------------------------------------------------


void randomize_sources_positions_(void)
{
  //std::random_device rd;
  std::default_random_engine reng(/*rd()*/RANDOM_SEED);
  std::uniform_real_distribution<double> uniform_dist(
      -MAX_SOURCE_DIST_FROM_ORIGIN, MAX_SOURCE_DIST_FROM_ORIGIN);

  for (int i=0; i<N_SOURCES; i++) {
    double const x = uniform_dist(reng), y = uniform_dist(reng), z = uniform_dist(reng);
    double const d = sqrt(x*x + y*y + z*z);
    if (d < CELL_RADIUS or d > MAX_SOURCE_DIST_FROM_ORIGIN) {
      i--;
      continue;
    } else {
      sources_positions_[i][0] = x;
      sources_positions_[i][1] = y;
      sources_positions_[i][2] = z;
  } }

#ifdef LOGGING_ENABLED

  // -- logging
  //
  std::ofstream ofs("sources-positions.dat");
  for (int i=0; i<N_SOURCES; i++) {
    ofs << sources_positions_[i][0] << " " 
        << sources_positions_[i][1] << " " 
        << sources_positions_[i][2] << std::endl;
  }

#endif
}


void randomize_sources_activities_(void)
{
  // -- times to onset
  //
  //std::random_device rd;
  std::default_random_engine reng(/*rd()*/RANDOM_SEED+1);
  std::uniform_real_distribution<double> uniform_dist(0,SOURCES_ACTIVITY_DURATION);

  for (int i=0; i<N_SOURCES; i++) {
    sources_activities_[i][0] = uniform_dist(reng);
  }

  // -- amplitude & duration (i.e., turn off times)
  //
  std::lognormal_distribution<double> lognorm_dist(SOURCES_AMPLITUDE_LOGNORM_MEAN,
                                                   SOURCES_AMPLITUDE_LOGNORM_STDDEV);
  std::exponential_distribution<double> exp_dist(1.0/SINGLE_SOURCE_MEAN_DURATION);

  for (int i=0; i<N_SOURCES; i++) {
    double const duration = exp_dist(reng);
    double const amplitude= lognorm_dist(reng);
    sources_activities_[i][1] = sources_activities_[i][0] + duration;
    sources_activities_[i][2] = amplitude;
  }

#ifdef LOGGING_ENABLED

  struct event_t { double tm; bool sw; };
  event_t events[2*N_SOURCES];
  for (unsigned int i=0; i<N_SOURCES; i++) {
    events[2*i  ].tm = sources_activities_[i][0];
    events[2*i  ].sw = sources_activities_[i][2];
    events[2*i+1].tm = sources_activities_[i][1];
    events[2*i+1].sw = 0.0;
  }
  std::sort(events, events+2*N_SOURCES,
      [](event_t const& a, event_t const& b){ return a.tm <= b.tm; });

  std::ofstream ofs2("sources-durations-amplitudes.dat");
  for (unsigned int i=0; i<N_SOURCES; i++) {
    ofs2 << sources_activities_[i][1] - sources_activities_[i][0] << ' ' 
         << sources_activities_[i][2] << std::endl;
  }

  std::ofstream ofs1("sources-timecourses.dat");
  for (unsigned int i=0; i<2*N_SOURCES; i++) {
    ofs1 << events[i].tm << ' ';
    for (int j=0; j<N_SOURCES; j++) {
      if (     sources_activities_[j][0] <= events[i].tm 
          and  events[i].tm <= sources_activities_[j][1]) {
        ofs1 << sources_activities_[j][2] << " ";
      } else {
        ofs1 << "0 ";
    } }
    ofs1 << std::endl;
  }

  // -- an alternative format:
  //
  //std::ofstream ofs1("sources-timecourses.dat");
  //for (double t=0; t<SOURCES_ACTIVITY_DURATION; t+=0.1) { // arbitrary timepoints
  //  ofs1 << t << " ";
  //  for (int i=0; i<N_SOURCES; i++) {
  //    if (sources_activities_[i][0] < t and t < sources_activities_[i][1]) {
  //      ofs1 << sources_activities_[i][2] << " ";
  //    } else {
  //      ofs1 << "0 ";
  //  } }
  //  ofs1 << std::endl;
  //}

#endif 
}


// x, y, z -- distances in three directions from the kernel center at (0,0,0)
//
inline double kernel_(double x, double y, double z, double t, double S, double D, double G)
{
  if (t > 0) {
    double const r2 = x*x + y*y + z*z;
    double const r  = sqrt(r2);
    double const B  = sqrt(G/D);
    double const Br = B*r;
    double const exp_Br = exp(Br);
    double const two_sqrt_Dt = 2.*sqrt(D*t);
    double const two_BDt = 2.*B*D*t;
    return -S * \
        (   (erf((r - two_BDt)/two_sqrt_Dt) + 1.)/exp_Br
          + (erf((r + two_BDt)/two_sqrt_Dt) - 1.)*exp_Br )
       /(8.*pi*D*r);
  } else {
    return 0.0;
  }
}


inline double kernel_dcoord__(double dcoord,
                              double x, double y, double z, double t, double S, double D, double G)
{
  if (t > 0) {
    double const r2 = x*x + y*y + z*z;
    double const r  = sqrt(r2);
    double const B  = sqrt(G/D);
    double const Br = B*r;
    double const sqrt_Dt = sqrt(D*t);
    double const two_sqrt_Dt = 2.*sqrt(D*t);
    double const sqrt_AD = sqrt(G*D);
    return 
      S* \
      (dcoord*(  2.
           + 2.*r*(B - exp((G*r)/sqrt_AD - r2/(4.*D*t) - G*t)/(sqrt_pi*sqrt_Dt))
           - (1. + Br)*erfc((r - 2.*sqrt_AD*t)/two_sqrt_Dt) 
           + exp(2.*Br)*(-erfc((r + 2.*sqrt_AD*t)/two_sqrt_Dt) + Br*erfc(sqrt(G*t) + r/two_sqrt_Dt)) ))
     /(8.*D*exp(Br)*pi*r*r2);
  } else {
    return 0.0;
  }
}
//
//
inline double kernel_dx_(double x, double y, double z, double t, double S, double D, double G)
{
  return kernel_dcoord__(x, x,y,z,t, S, D,G);
}
//
//
inline double kernel_dy_(double x, double y, double z, double t, double S, double D, double G)
{
  return kernel_dcoord__(y, x,y,z,t, S, D,G);
}
//
//
inline double kernel_dz_(double x, double y, double z, double t, double S, double D, double G)
{
  return kernel_dcoord__(z, x,y,z,t, S, D,G);
}


inline double kernel_dt_(double x, double y, double z, double t, double S, double D, double G)
{
  if (t > 0) {
    double const r2 = x*x + y*y + z*z;
    double const Dt = D*t;
    return S*exp(-r2/(4.*Dt) - G*t)/(8.*pi_pow_3over2*pow(Dt,1.5));
  } else {
    return 0.0;
  }
}


template <class TKernelF>
inline double sum_kernels_(TKernelF kernel, double x, double y, double z, double t)
{
  double sum = 0.0;
  for (unsigned int i=0u; i<N_SOURCES; i++) {
    double const t_begin = sources_activities_[i][0], 
                 t_end   = sources_activities_[i][1];
    if (t < t_begin) {
      continue;
    } else {
      double const amplitude = sources_activities_[i][2];
      double const dx = x - sources_positions_[i][0];
      double const dy = y - sources_positions_[i][1];
      double const dz = z - sources_positions_[i][2];
      if (t < t_end) {
        sum +=   kernel(dx,dy,dz, t - t_begin,
                       amplitude, SOURCES_ALL_DIFFUSIVITY, SOURCES_ALL_DEGRADATION)
               - kernel(dx,dy,dz, nearly_zero,
                       amplitude, SOURCES_ALL_DIFFUSIVITY, SOURCES_ALL_DEGRADATION);
      } else {
        sum +=   kernel(dx,dy,dz, t - t_begin + nearly_zero,
                       amplitude, SOURCES_ALL_DIFFUSIVITY, SOURCES_ALL_DEGRADATION)
               - kernel(dx,dy,dz, t - t_begin - (t_end - t_begin),
                       amplitude, SOURCES_ALL_DIFFUSIVITY, SOURCES_ALL_DEGRADATION);
    } }
  }
  return sum;
}

